/******************************************************************/
/*                  Digitrak Flight Analyzer (tm)                 */
/*             Analysis of Consumer Rocketry Flight Data          */
/*                      Copyright 12 May 1996                     */
/*                         Larry Curcio                           */
/*                Free Not-For-Profit Distribution                */
/*                    All Other Rights Reserved                   */
/*                      Beta Test Version                         */
/******************************************************************/

/* BIG C (Or C++--) : C++ code written in C  */
/* Object Orientation Simplifies Overloading of Alternate Methods */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>


#define ABS  abs
#define REAL float

#define NINT int
#define VOID void
#define CHAR char
#define UCHAR unsigned char
#define CBOOL UCHAR
#define STATIC static

#define EPSILON (REAL) 0.0001
#define MISSING (REAL) -999e35

/* Brand Switch For Data Interpretation */

#define CAMBRIDGE 1
#define CAMBRIDGEP 2
#define ACCELERATION 3
#define ADEPT 4

#define CAMBRIDGE_PREAMBLE_SIZE 12
#define CAMBRIDGE_PRINT_PREAMBLE_SIZE 22
#define CAMBRIDGE_INTERVAL_INDEX 2
#define CAMBRIDGE_TICK_SIZE ((REAL) 0.016384)

#define RHO0 (REAL) 1.2929
#define KELVIN0 (REAL) 273.15

#define PI (REAL) 3.14159265358979

/* GRAVITATIONAL CONSTANT IN m/sec^2 */
#define G (REAL) 9.805

#define MTOF ((REAL) 3.28084)
#define OZTOKG ((REAL) 0.02835)

#define GENERIC 1
#define TURBODOS 2
#define UNIX 3
#define VCPLUSPLUS 4
#define MAC 5


#define METRIC 1
#define ENGLISH 2
#define COMBO 3

#define SQR(X) X*X

#ifndef SMOOTHING_CONSTANT
#define SMOOTHING_CONSTANT ((REAL) 0.55)
#endif

#define SYSTEM_TYPE GENERIC


/* Brand Switch For Data Interpretation */

#ifndef TRUE
#define TRUE (CBOOL)  1
#define FALSE (CBOOL) 0
#endif

#define MAX_STRING 128

#ifndef STDOUT
#define STDOUT stdout
#endif


#ifndef DISPLAY
#define DISPLAY printf
#define DISPLAYF fprintf
#endif

#ifndef ERROR_MESSAGE
#define ERROR_MESSAGE(X) fprintf(stderr, X, Y)
#endif

#ifndef ERROR_MESSAGE1
#define ERROR_MESSAGE1(X, Y) fprintf(stderr, X, Y)
#endif

#ifndef LABEL_SIZE
#define LABEL_SIZE 32
#endif


#ifndef CD_TABLE_SIZE
#define CD_TABLE_SIZE 1000
#endif

#ifndef MAX_SIZE
#define MAX_SIZE 500
#endif

#ifndef THRUST_SIZE
#define THRUST_SIZE 250
#endif

#ifndef MAX_STAGE
#define MAX_STAGE 1
#endif

#ifdef NOT_INTERACTIVE
#define CLEAR_SCREEN /* NOT INTERACTIVE */
#endif

#ifndef CLEAR_SCREEN
#if SYSTEM_TYPE == UNIX
#define CLEAR_SCREEN system("clear");
#elif (SYSTEM_TYPE == TURBODOS)
#define CLEAR_SCREEN clrscr()
#else
#define CLEAR_SCREEN DISPLAY("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
#endif
#endif


/**************************************************************************/
/*                        .h File Stuff                                   */
/**************************************************************************/

#define STAGE_CLASS_HEADER \
NINT nStageNumber; \
NINT nMaxSize; \
NINT nMachCorrectSwitch; \
NINT nCdTableSize; \
NINT nDataKind; \
NINT nSizeCoastPhase; \
NINT nStartCoastPhase; \
REAL rMaxCd; \
REAL rDeltat; \
REAL rMaxDiamInput; \
REAL rMaxDiam; \
REAL rTotalImpulse; \
REAL rMassPropInput; \
REAL rMassProp; \
REAL rLaunchMassInput; \
REAL rLaunchMass; \
REAL rIsp; \
REAL rVe; \
REAL rAx; \
REAL rDeltaT; \
REAL rVMax; \
REAL rAMax; \
REAL rAMin; \
REAL rLastV; \
REAL rThrust[MAX_SIZE]; \
REAL rV[MAX_SIZE]; \
REAL rA[MAX_SIZE]; \
REAL rAlt[MAX_SIZE]; \
REAL rCd[MAX_SIZE]; \
REAL rK[MAX_SIZE]; \
REAL rCdTable[CD_TABLE_SIZE]; \
CHAR caInputFileName[MAX_STRING]; \
CHAR caInputCdCurveName[MAX_STRING]; \
CHAR caInputFCurveName[MAX_STRING]; \
CHAR caOutputFCurveName[MAX_STRING]; \
CHAR caOutputCdCurveName[MAX_STRING]; \
struct _clconfig * ClpConfig; \
VOID (* vSaveFlightFile)(struct _clstage * ClpThis, FILE * ofp); \
VOID (* vSaveCdCurve)(struct _clstage *  ClpStage); \
VOID (* vSaveFCurve)(struct _clstage *  ClpStage); \
CBOOL (* bReadInputData) (struct _clstage * ClpStage); \
CBOOL (* bReadInputKeyboard) (struct _clstage * ClpStage); \
REAL (* rRho)(struct _clstage * ClpThis, REAL H); \
NINT (* nFindBurnout)(struct _clstage * ClpThis); \
VOID (* vGetVMaxAMaxAMin)(struct _clstage * ClpThis); \
VOID ( * vFillinDataBlanks)(struct _clstage * ClpThis, REAL rA0, REAL rV0, \
				 REAL rAlt0); \
VOID ( * vConvolve)(struct _clstage * ClpThis, REAL *rArray, NINT nSize); \
VOID ( * vSmooth)(struct _clstage * ClpThis, REAL *rArray, NINT nSize); \
VOID ( * vAssertMaximumLowVCd)(struct _clstage * ClpThis); \
VOID ( * vMakeShapeConstantsAndCds)(struct _clstage * ClpThis); \
REAL ( * rMakeSpaceVelocities)(struct _clstage * ClpThis, REAL rSpaceV[], \
				 REAL rM[]); \
REAL ( * rMakeThrustCurve)(struct _clstage * ClpThis, REAL rM[], \
					       CBOOL bInitialize); \
CBOOL (* bAnalyzeStage)(struct _clstage * ClpThis, NINT nNumStage); \
VOID (* vReadCdCurve)(struct _clstage * ClpThis); \
VOID (* vReadThrustCurve)(struct _clstage * ClpThis); \
VOID (* vPrintoutStage) (struct _clstage * Clpthis, NINT nNumStage, \
				    FILE * ofp); \
VOID (* vDestructor) (struct _clstage * ClpThis);


/* Stage Class Pointer */
/* Note that lower stages must be analyzed differently from upper stages */
/* Even in lower stages, a number of options exist */
/* Hence the function pointers at the stage level */

typedef struct _clstage {
STAGE_CLASS_HEADER
} Stage_Cl, * Stage_Clp;



#define CONFIG_CLASS_HEADER \
NINT nNumStages; \
NINT nUnits; \
NINT nBrandOfInstrument; \
NINT nNoseSwitch; \
REAL rRho0; \
REAL rAlt0Input; \
REAL rAlt0; \
REAL rTempInput; \
REAL rTempF; \
REAL rTempC; \
REAL rTempK; \
REAL rMaxDiameterFactor; \
REAL rMassPropFactor; \
REAL rLaunchMassFactor; \
REAL rAlt0Factor; \
CHAR caOutputFlightFileName[MAX_STRING]; \
CHAR caTempLabel[LABEL_SIZE]; \
CHAR caMaxDiameterLabel[LABEL_SIZE]; \
CHAR caMassPropLabel[LABEL_SIZE]; \
CHAR caLaunchMassLabel[LABEL_SIZE]; \
CHAR caAlt0Label[LABEL_SIZE]; \
Stage_Cl ClStage[MAX_STAGE]; \
Stage_Clp ClpStage[MAX_STAGE]; \
VOID (* vSaveFlightFile)(struct _clconfig * ClpThis); \
CBOOL (* bInputConfiguration)(struct _clconfig * ClpThis); \
VOID (* vAnalyzeConfiguration)(struct _clconfig * ClpThis); \
VOID (* vPrintoutConfiguration)(struct _clconfig * ClpThis, FILE * ofp); \
VOID (* vSaveConfigurationData)(struct _clconfig * ClpThis); \
VOID (* vDestructor) (struct _clconfig * ClpThis);

/* Configuration Class Pointer */
typedef struct _clconfig{
CONFIG_CLASS_HEADER
} Config_Cl, *Config_Clp;

/**************************************************************************/
/*                  Begin Executable Code                                 */
/**************************************************************************/

VOID Stage_Cl__vDoNothing(Stage_Clp ClpThis)
{
ClpThis = ClpThis;
/* This Function Does Absolutely Nothing */
return;
}
/***************************************************************************/
int EditNum(char * instring)
/* *** EDIT FOR VALID NUMERIC *** */
{
int  numpt=0, numblank=0, numneg=0, notblank=0;

while((*instring !='\0') &&(*instring != '\r'))
  {
  if((*instring < '0') || (*instring >'9'))
    {

    if(*instring == ' ')
      {
      if(notblank) ++numblank;
      else numblank=1;
      }
    else  /* (*instring == ' ') */
    if(*instring == '-')
      {
      ++numneg;
      if(notblank)return(FALSE);
      }
    else if(*instring == '.')++numpt;

    else return(FALSE); /* non-numeric character found */

    notblank=1;
    }
  else notblank=1;
  ++instring;
  }
  if(numblank > 2)return(FALSE);
  if(numpt > 1)   return(FALSE);
  if(numneg > 1)  return(FALSE);
  return(TRUE);
}
/***************************************************************************/

float GetVal(char * cp)

/* *** ACCEPT NUMERIC FROM KEYBOARD, EDIT FOR VALIDITY, AND CONVERT *** */
{
  float res;

  if(!EditNum(cp))return(MISSING);
  sscanf(cp, "%f", &res);
  return(res);
}

/***************************************************************************/
VOID GetLine(CBOOL upcase, CHAR * cp)

/* *** ACCEPT A LINE FORM THE KEYBOARD *** */
{
  int c;

  for (;;)
    {
    c = getchar();
    if ((c >= 'a') && (c <= 'z') && (upcase))
	c = c + 'A' - 'a';
    if (c == EOF || c == '\n')
      {
      *cp = 0;
      return;
      }
    *cp++ = c;
    }
}

/***************************************************************************/
float Prompt(char * instring, float x)
/* *** PROMPT USER FOR NUMERIC VALUE AND ACCEPT *** */
/* *** NUMERIC VALUE IS FLOATING POINT, BUT MORE VERSATILE WITH CASTS *** */
{
#ifndef MISSING
#define MISSING -999e35
#endif
int go_on;
float y;
char response[MAX_STRING];

go_on=TRUE;

while (go_on)
  {
  y=x;
  printf("%s (%9.4f) ",instring, x);
  GetLine((int) TRUE, response);
  if( response[0] != '\0') y = GetVal(response);
  if (y > MISSING)
#if(SYSTEM_TYPE != TURBODOS)
     {
     go_on=FALSE;
     }
#else
    {
    go_on=FALSE;
    gotoxy(1,wherey()-1);
    clreol();
    printf("%s (%9.4f) \n\r",instring, y);
    }
  else
    {
    gotoxy(1,wherey()-1);
    clreol();
    }
#endif
  }

return(y);
}
/***************************************************************************/
CHAR GetReply(legend)

/* *** ACCEPT SINGLE CHARACTER REPLY AND PUT INTO UPPER CASE *** */

char legend[];
{
char reply;
char linein[10];
DISPLAY("%s ", legend);

do{
  GetLine(1, linein);
  reply = linein[0];
  }while (strlen(linein) != 1);

#if(SYSTEM_TYPE == TURBODOS)
gotoxy(1,wherey()-1);
printf("%s %c \r", legend, reply);
#endif

return(reply);
}
/**************************************************************************/
VOID vBanner()
{
CHAR capCharArray[16];

CLEAR_SCREEN;

DISPLAY("\n\n\n\n\n\n\n");

DISPLAY(
"                         Digitrak Flight Analyzer (tm)\n\n");
DISPLAY(
"                   Analysis of Consumer Rocketry Flight Data\n");
DISPLAY(
"                             Copyright 12 May 1996\n");
DISPLAY(
"                                 Larry Curcio\n");
DISPLAY("\n");

DISPLAY(
"                            * Beta Test Version *     \n");
DISPLAY("\n");

DISPLAY(
"                       Free Not-For-Profit Distribution\n");
DISPLAY(
"                           All Other Rights Reserved\n");


GetLine(FALSE, capCharArray);

return;
}
/**************************************************************************/
VOID Stage_Cl__vGetVMaxAMaxAMin(Stage_Clp ClpThis) {
NINT i;
REAL rVMax, rAMax, rAMin;

rVMax = rAMax = rAMin = (REAL) 0.0;


for(i=0; i< ClpThis->nMaxSize; i++){
  if(ClpThis->rV[i] > rVMax) {
    rVMax = ClpThis->rV[i];
  }

  if(ClpThis->rA[i] > rAMax) {
    rAMax = ClpThis->rA[i];
  }

 if(ClpThis->rA[i] < rAMin) {
    rAMin = ClpThis->rA[i];
  }

}

ClpThis->rVMax = rVMax;
ClpThis->rAMax = rAMax;
ClpThis->rAMin = rAMin;

return;
}
/**************************************************************************/
REAL Stage_Cl__rRho(Stage_Clp ClpThis, REAL H)
{
REAL rWork;
 H += ((Config_Clp) ClpThis->ClpConfig)->rAlt0;
 rWork = exp(H * ((REAL) -.00008187+ H * ((REAL)-3.085e-9 + H *
		  (REAL)(3.975e-14))));

 return(RHO0*KELVIN0 * rWork / ((Config_Clp) ClpThis->ClpConfig)->rTempK);
}
/**************************************************************************/
NINT Stage_Cl__nFindBurnout(Stage_Clp ClpThis)
{
/* Finds Approximate Start of Coast Phase as Last Minimum Acceleration */
/* @@@ Modify for Multiple Stages */
NINT i, nIMinimum;
REAL rWork = -MISSING, *rA;

rA = ClpThis->rA;

for(i = (3* ClpThis->nMaxSize) / 4; i >= 0; i--) {
   if(rA[i] <= rWork) {
     rWork = rA[i];
     nIMinimum = i;
   }
}

ClpThis->nStartCoastPhase = nIMinimum;
return(nIMinimum);
}
/**************************************************************************/
VOID Stage_Cl__vMakeAccsAndAlts(Stage_Clp ClpThis,
					  REAL rA0,
					  REAL rV0,
					  REAL rAlt0)
{

/*        **** Cambridge FillinDataBlanks Routine ****          */
/*   **** Acceleration & Altitude from Velocity & Time ****     */
/* Compute Accelerations And Altitudes From Velocities */
/* Note That rV0 is Ignored. It's There to Simplify Overloading */

NINT i;
REAL * rA, * rV, * rAlt;
REAL rAMean, rHalfDeltaT, rDeltaT;

rA = ClpThis->rA;
rV = ClpThis->rV;
rAlt = ClpThis->rAlt;
rDeltaT = ClpThis->rDeltaT;
rHalfDeltaT = (REAL) 0.5 * ClpThis->rDeltaT;

rV[0] = rV0;
rA[0] = rA0;
rAlt[0] = rAlt0;


/* Velocity curve doesn't have a lot of twists */
/* Therefore it's a good candidate for smoothing */

ClpThis->vConvolve(ClpThis, rV, ClpThis->nMaxSize);

for(i=1; i < ClpThis->nMaxSize; i++) {
   /* Cambridge acceleration integration is rectangular */
    rA[i] = (rV[i] - rV[i-1])/rDeltaT;
}

for(i=1; i < ClpThis->nMaxSize; i++) {
   /* Cambridge Altitudes are rectangularly integrated. Recompute */
   rV[i] = rV[i-1] + (rA[i] + rA[i-1]) * rHalfDeltaT;
   /* And Do Altitudes While We're At It */
   rAlt[i] = rAlt[i-1] + (rV[i] + rV[i-1]) * rHalfDeltaT;
}

return;
}
/**************************************************************************/
VOID Stage_Cl__vMakeVelsAndAccs(Stage_Clp ClpThis,
					  REAL rA0,
					  REAL rV0,
					  REAL rAlt0)
{

/*        **** Adept FillinDataBlanks Routine ****                  */
/*  **** Velocity and Acceleration from Altitude & Time ****        */
/*       **** Adept Data Requires Serious Smoothing ****            */

/*        **** Not Available in Beta Test Version ****              */


NINT i;
REAL * rA, * rV, * rAlt;
REAL rDeltaT, rHalfDeltaT;


rA = ClpThis->rA;
rV = ClpThis->rV;
rAlt = ClpThis->rAlt;
rDeltaT = ClpThis->rDeltaT;
rHalfDeltaT = (REAL) 0.5 * rDeltaT;

rV[0] = rV0;
rA[0] = rA0;
rAlt[0] = rAlt0;

ClpThis->vConvolve(ClpThis, rAlt, ClpThis->nMaxSize);
ClpThis->vConvolve(ClpThis, rAlt, ClpThis->nMaxSize);


for(i=1; i < ClpThis->nMaxSize; i++) {
   rV[i] = (rAlt[i] - rAlt[i-1])/rDeltaT;
}

/* ClpThis->vSmooth(ClpThis, rV, ClpThis->nMaxSize); */


for(i=1; i < ClpThis->nMaxSize; i++) {
   rAlt[i] = rAlt[i-1] + (rV[i] + rV[i-1]) * rHalfDeltaT;
   rA[i] = (rV[i] - rV[i-1])/rDeltaT;
}

return;
}
/**************************************************************************/
VOID Stage_Cl__vMakeVelsAndAlts(Stage_Clp ClpThis,
					  REAL rA0,
					  REAL rV0,
					  REAL rAlt0)
{

/*     **** HomeMade Accelerometer FillinDataBlanks Routine ****      */
/*        **** Velocities & Altitdudes from Accelerations ****       */

NINT i;
REAL * rA, * rV, * rAlt;
REAL rHalfDeltaT;


rA = ClpThis->rA;
rV = ClpThis->rV;
rAlt = ClpThis->rAlt;
rHalfDeltaT = (REAL) 0.5 * ClpThis->rDeltaT;

rV[0] = rV0;
rA[0] = rA0;
rAlt[0] = rAlt0;
for(i=1; i < ClpThis->nMaxSize; i++) {
   rV[i] = rA[i-1] + (rA[i] + rA[i-1]) * rHalfDeltaT;
   rAlt[i] = rAlt[i-1] + (rV[i] + rV[i-1]) * rHalfDeltaT;
}
return;
}
/**************************************************************************/
VOID Stage_Cl__vAssertMaximumLowVCd(Stage_Clp ClpStage)
{
NINT i;
/* 100 m/sec used as arbitrary low-velocity limit */

for(i=0; i < 100; i++) {
   if(ClpStage->rCdTable[i] > ClpStage->rMaxCd) {
     ClpStage->rCdTable[i] = ClpStage->rMaxCd;
   }
}
return;
}
/**************************************************************************/
VOID Stage_Cl__vExponentialSmoothing(Stage_Clp ClpThis, REAL *rArray,
NINT nSize)
{
NINT i;
REAL rWork, rWorkLast;

/* This funtion smooths a curve by exponential smoothing */
ClpThis = ClpThis;
rWork = SMOOTHING_CONSTANT;
rWorkLast = (REAL) 1.0 - rWork;

		       /* BackCasting */

for (i = nSize - (NINT) 3; i > 0; i--) {

    rArray[i] = rWorkLast * ((REAL) 2.0 * rArray[i+1] - rArray[i+2])
			+ rWork * rArray[i];

}

/*
for (i=2; i < nSize - (NINT) 1; i++) {
    rArray[i] = rWorkLast * ((REAL) 2.0 * rArray[i-1] - rArray[i-2])
			+ rWork * rArray[i];
}
*/

return;
}
/**************************************************************************/
VOID Stage_Cl__vConvolve(Stage_Clp ClpThis, REAL *rArray, NINT nSize)
{
NINT i;
REAL rWork, rWorkLast;
/* This funtion smooths a curve by averaging with nearest neighbors */
ClpThis = ClpThis;
rWork = rArray[0];
for (i=1; i < nSize -(NINT) 1; i++) {
    rWorkLast = rWork;
    rWork = rArray[i];
    rArray[i] = (rWork + rWorkLast + rArray[i+1]) / (REAL) 3.0;
}
rArray[nSize-1] = (rWork + rArray[nSize-1]) / (REAL) 2.0;

return;
}
/**************************************************************************/
REAL GetMach1(REAL Temp)
/* *** FIND MACH 1 FROM TEMPERATURE *** */
{
#ifndef MACH1AT0
#define MACH1AT0 331.45
#endif
return(MACH1AT0*sqrt(Temp/273.15));
}
/***************************************************************************/
VOID MachMaker(REAL cd[], REAL Temp, NINT nosesw)
{
/*COMPILE OPTION*/
/*
This routine accounts for drag divergence in the manner described by
Mandell, Caporaso, and Bengen in _TOPICS IN ADVANCED MODEL ROCKETRY_,
first edition. The authors use a critical mach number of 0.9. I have
converted their model to use a critical mach number of 0.8, but I have
given the user a choice of either value.

To choose a critical mach number, #define MCRIT as either 8 (for 0.8) or 9
(for 0.9). Since some c compilers running under UNIX cannot understand a
#if comparison to a decimal constant, I had to use whole numbers instead
of the values themselves.
*/
#define MCRIT 8


/* DON'T CHANGE THIS */
#define SHARP 1

REAL MACH1, MACHPT9, MACHPT8, fwork, fwork1, machnum;
NINT IMCRIT, IMACHPT9, IMACHPT8, IMACH1PT05, IMACH1PT2, i, smaller;


MACH1 = GetMach1(Temp);
MACHPT9 = 0.9*MACH1;
IMACHPT9 = (NINT) MACHPT9;
MACHPT8 = 0.8*MACH1;
IMACHPT8 = (NINT) MACHPT8;

if ((REAL) IMACHPT9 < MACHPT9) IMACHPT9++;
if ((REAL) IMACHPT8 < MACHPT8) IMACHPT8++;

IMACH1PT05 = (NINT) (MACH1*1.05 + .999999);
IMACH1PT2 =  (NINT) (MACH1*1.2  + .999999);

#if (MCRIT == 9)
  IMCRIT=IMACHPT9;
  if(IMACHPT9 > CD_TABLE_SIZE) return;
  fwork = cd[IMACHPT9-1];
  fwork1=(REAL) IMACHPT9;
#else
  IMCRIT=IMACHPT8;
  if(IMACHPT8 > CD_TABLE_SIZE) return;
  fwork = cd[IMACHPT8-1];
  fwork1=(REAL) IMACHPT8;
#endif

if(nosesw == SHARP)
  {
      /* Sharp Nosed Vehicles */

  if(IMACH1PT05 < CD_TABLE_SIZE) smaller=IMACH1PT05;
  else smaller = CD_TABLE_SIZE;

  for (i=IMCRIT; i<smaller; i++)
      {
      machnum=fwork1/MACH1;
      fwork1 +=1.0;
      /* (0.9 =< M =< 1.05)  */
#if(MCRIT == 9)
      cd[i] = fwork*(1.0 + 35.5*(machnum-0.9)*(machnum-0.9));
#else
      cd[i] = fwork*(1.0 + 12.78*(machnum-0.8)*(machnum-0.8));
#endif

      }
  if(smaller < CD_TABLE_SIZE) for (i=smaller; i< CD_TABLE_SIZE; i++)
      {
      machnum=fwork1/MACH1;
      fwork1 +=1.0;
      /*  1.05 =< M =< 2.0  */
      cd[i]=fwork*(1.27 + 0.53*exp(-5.2*(machnum-1.05)));
      }
  }
else

  {
	   /* Round-Nosed Vehicles */
  if(IMACH1PT2 < CD_TABLE_SIZE) smaller=IMACH1PT2;
  else smaller = CD_TABLE_SIZE;
for (i=IMCRIT; i<smaller; i++)
      {
      machnum=fwork1/MACH1;
      fwork1 +=1.0;
      /* (0.9 =< M =< 1.2) */
#if(MCRIT == 9)
      cd[i] = fwork * (1.0 + 4.88*exp(1.1*log(machnum-0.9)));
#else
      cd[i] = fwork * (1.0 + 3.556208468*exp(1.1*log(machnum-0.8)));
#endif
      }
  if(smaller < CD_TABLE_SIZE) for (i=smaller; i<CD_TABLE_SIZE; i++)
      {
      machnum=fwork1/MACH1;
      fwork1 +=1.0;
      /* 1.2 =< M =< 2.0 */
      cd[i] = fwork * (2.0 + 0.3*exp(-5.75*(machnum-1.2)));
      }
  }
  return;
}
/***************************************************************************/
VOID Stage_Cl__vMakeShapeConstantsAndCds(Stage_Clp ClpThis)
{
/* Compute Shape Constants and Cd's in Coast Phase Only */

REAL * rCdBoost, * rKBoost, * rK, * rA, * rAlt, * rV, * rCd, * rCdTable;
REAL * rVBoost, * rAltBoost;
REAL rAx, rMBurnout, rDeltT, rWork;
NINT nMaxSize, i, j, nLastV;
REAL rVelocity;
Config_Clp ClpConfig;

nMaxSize = ClpThis->nMaxSize - ClpThis->nStartCoastPhase;
rA = &ClpThis->rA[ClpThis->nStartCoastPhase];
rAlt = &ClpThis->rAlt[ClpThis->nStartCoastPhase];
rV = &ClpThis->rV[ClpThis->nStartCoastPhase];
rK = &ClpThis->rK[ClpThis->nStartCoastPhase];
rVBoost =ClpThis->rV;
rKBoost = ClpThis->rK;
rCdBoost = ClpThis->rCd;
/* rAltBoost = ClpThis->rAlt; */
rCd = &ClpThis->rCd[ClpThis->nStartCoastPhase];
rCdTable = ClpThis->rCdTable;
rAx = ClpThis->rAx;
rMBurnout = ClpThis->rLaunchMass - ClpThis->rMassProp;

for(i= 0 ; i < nMaxSize; i++) {

   rWork = (rV[i] * rV[i])*ClpThis->rRho(ClpThis, rAlt[i]);

   if(rWork == (REAL) 0.0) {
     DISPLAY("Zero Divide in MakeShapeConstantsAndCds: i= %i\n", i);
     exit(0);
   }

   rK[i] =  - (rA[i] + G) * rMBurnout* RHO0 /rWork;


   rCd[i] = - (REAL) 2.0 * (rA[i] + G) * rMBurnout/(rWork * rAx);

   if(rCd[i] < (REAL) 0.0) {
      rCd[i] = rK[i] = (REAL) 0.0;
   }

}

/* Fill In Boost Phase Here For Thrust Curve Calculation */

for (i = 0; i < ClpThis->nStartCoastPhase; i++) {
    rVelocity = rVBoost[i];
    for (j = 0; j < nMaxSize; j++) {

	if(rVelocity >= rV[j]) {

	  rKBoost[i] = rK[j] + (rK[j - 1] - rK[j]) *
			    (rVelocity -rV[j])/(rV[j - 1] - rV[j]);

	  rCdBoost[i] = (rCd[j] + (rCd[j - 1] - rCd[j]) *
	       (rVelocity -rV[j])/(rV[j - 1] - rV[j]));

	  break;
	} /* If */

    }/* For j */

} /* For i */

rVelocity = (REAL) 0.0;
/* Make CdTable By Velocity */



for (i = 1; i < ClpThis->nCdTableSize; i++) {

    rVelocity += 1.0;

    for (j = 0; j < nMaxSize;  j++) {

	if(rVelocity >= rV[j]) {

	  if(j == 0 ) {
	    /* Extrapolate Drag Coefficients */
	    /* Should Be Mach Adjusted Later */
	     rCdTable[i] = rCdTable[i - 1];
	    break;
	  } /* If J */
	  if (j > 0) {
	    rCdTable[i] = rCd[j] - (rCd[j - 1] - rCd[j]) *
			     (rVelocity -rV[j])/(rV[j - 1] - rV[j]);
	  }
	  else {
	    rCdTable[i] = rCd[0];
	  }

	  break;
	} /* If rVelocity */

    }/* For j */

} /* For i */

/* Cd at 0 velocity is the sound of 1 hand clapping anyway */


/* Propagate Backward: Cd of Minimum Coast Phase Velocity */
ClpThis->rLastV = rVBoost[ClpThis->nMaxSize -1];
nLastV = (NINT) (ClpThis->rLastV) + 2;
if(((REAL) nLastV - 2) == ClpThis->rLastV) {
  nLastV--;
}
/* rWork now Cd for lowest velocity */
rWork =  rCdTable[nLastV];

for(i = nLastV; i>=0; i--) {

  rCdTable[i] = rWork;

}

ClpThis->vConvolve(ClpThis, &rCdTable[nLastV], ClpThis->nCdTableSize - nLastV);
ClpThis->vConvolve(ClpThis, &rCdTable[nLastV], ClpThis->nCdTableSize - nLastV);
/* ClpThis->vConvolve(ClpThis, &rCdTable[nLastV], ClpThis->nCdTableSize - nLastV); */

/* Stage_Cl__vAssertMaximumLowVCd(ClpThis); */

ClpConfig =(Config_Clp) (ClpThis->ClpConfig);

if(ClpThis->nMachCorrectSwitch == 1) {
  MachMaker(rCdTable, ClpConfig->rTempK, (NINT) ClpConfig->nNoseSwitch);
}
return;
}
/**************************************************************************/
REAL Stage_Cl__rMakeSpaceVelocities(Stage_Clp ClpThis,
				      REAL rSpaceV[],
				      REAL rM[])
{
/* Infer Equivalent Interval Velocity Changes in Outer Space */
/* This version uses empirical shape constants - unsmoothed */

REAL * rA, * rAlt, *rV, rSpaceAcc, rOldSpaceAcc, rDeltaT, rHalfDeltaT;
REAL rVMax, * rCd, rADrag;
/* REAL rSpaceV[MAX_STAGE]; */
NINT i;

/* Unfold Arrays */
rCd = ClpThis->rCdTable;
rA = ClpThis->rA;
rAlt = ClpThis->rAlt;
rV = ClpThis->rV;
rDeltaT = ClpThis->rDeltaT;
rHalfDeltaT = (REAL) 0.5 * rDeltaT;

rOldSpaceAcc = (REAL) 0.0;
/* The Zeroeth Element May be Filled in Previously when Extended */

if(rA[0] != (REAL) 0.0) {
   rADrag = rCd[(NINT) (rV[0] + (REAL) 0.5)] * rV[0]* rV[0] * RHO0 * ClpThis->rAx /
		  ((REAL) 2.0 * ClpThis->rRho(ClpThis, rAlt[0])*rM[0]);

   rOldSpaceAcc = rA[0] + G + rADrag;
}

rSpaceV[0] = (REAL) 0.0;

for(i=1; i < ClpThis->nStartCoastPhase; i++) {

   /* Add Gravity and Drag Accelerations Back */
   rADrag = rCd[(NINT) (rV[i] + (REAL) 0.5)] * rV[i]* rV[i] * RHO0 * ClpThis->rAx /
		  ((REAL) 2.0 * ClpThis->rRho(ClpThis, rAlt[i])*rM[i]);

   rSpaceAcc = rA[i] + G + rADrag;

   rSpaceV[i] = rVMax = rSpaceV[i - 1] +
			 (rSpaceAcc + rOldSpaceAcc) * rHalfDeltaT;


   rOldSpaceAcc = rSpaceAcc;

}
for(i=1; i < ClpThis->nStartCoastPhase; i++) {
  if(rSpaceV[i] > rVMax) {
     rVMax = rSpaceV[i];
  }
}
return(rVMax);
}

/**************************************************************************/
REAL Stage_Cl__rMakeThrustCurve(Stage_Clp ClpThis, REAL rM[],
						  CBOOL bInitialize)
{
REAL rDeltaV, rMBurnout, rTotalImpulse, rRunningImpulse, rOldImpulse;
REAL rM1, rM2, rMassProp, rLaunchMass, rVe;
REAL rSpaceV[MAX_SIZE], *rThrust, rMeanThrust, rDeltaT;
REAL rDeltaMass, rRunningMass;
NINT nMaxSize, i;

rMBurnout = ClpThis->rLaunchMass - ClpThis->rMassProp;
rLaunchMass = ClpThis->rLaunchMass;

nMaxSize = ClpThis->nStartCoastPhase;
rMassProp = ClpThis->rMassProp;
rDeltaT = ClpThis->rDeltaT;
rDeltaMass = rMassProp/((REAL) (ClpThis->nStartCoastPhase - 1));
rRunningMass = rMassProp;

 /* First Guess at Mass Distribution is Proportional to Time */
if(bInitialize == TRUE) {
  for (i=0; i < ClpThis->nStartCoastPhase; i++) {

     rM[i] = rRunningMass + rMBurnout;
     rRunningMass -= rDeltaMass;

  }
}

rDeltaV = -MISSING;


   /* Create an Array with Equivalent Outer Space Velocities */
   rDeltaV = ClpThis->rMakeSpaceVelocities(ClpThis, rSpaceV, rM);

   /* Compute Effective Exhaust Velocity of Stage */

   ClpThis->rVe = rVe = rDeltaV/log(ClpThis->rLaunchMass/rMBurnout);

   /* Compute Specific Impulse of Stage */
   ClpThis->rIsp = ClpThis->rVe / G;

   /* Compute Total Impulse of Stage */
   ClpThis->rTotalImpulse = rTotalImpulse = ClpThis->rVe * ClpThis->rMassProp;

   rM1 = rM[0] = rM2 = ClpThis->rLaunchMass;
   rRunningImpulse = rOldImpulse = (REAL) 0.0;
   rThrust = ClpThis->rThrust;
   rThrust[0] = (REAL) 2.0 * rLaunchMass * ClpThis->rV[0]/rDeltaT;

   for(i = 1; i < nMaxSize; i++) {
      /* From Rocket Equation */
      rM2 = rM[i] = rM1 * exp((rSpaceV[i-1] - rSpaceV[i]) / ClpThis->rVe);

      /* Constant Ve Approximation: Impulse Proportional to Propellant */
      rOldImpulse = rRunningImpulse;
      /*rRunningImpulse = rTotalImpulse * (rLaunchMass - rM2) / rMassProp;*/
       rRunningImpulse = rVe * (rLaunchMass - rM2);

      /* Make Thrust Fit Trapezoidal Approximation */
      /*
      rMeanThrust = (rRunningImpulse - rOldImpulse) / rDeltaT;
      rThrust[i] = (rMeanThrust + rMeanThrust) - rThrust[i-1];
      */
      rThrust[i] = (rRunningImpulse - rOldImpulse) / rDeltaT;

      rM1 = rM2;
   }
   if(rThrust[i-1] < 0.0) {
     rTotalImpulse = - (REAL)abs(rTotalImpulse);
     for( i = 0; i < ClpThis->nStartCoastPhase; i++) {
       if (rThrust[i] < 0.0) {
	  ClpThis->nStartCoastPhase = i;
	  break;
       }

     }
   }
 return (rTotalImpulse);
}
/**************************************************************************/
CBOOL Stage_Cl__bAnalyzeStage(Stage_Clp ClpThis, NINT nNumStage)
{
REAL rA0, rAlt0, rV0;
Stage_Clp ClpLastStage;
REAL rTotalImpulse, rOldTotal;
REAL rM[THRUST_SIZE];
NINT i;
STATIC CBOOL bRepeat = FALSE;

rA0 = rAlt0 = rV0 = (REAL) 0.0;

if(nNumStage != 0) {
  ClpLastStage = ((Config_Clp)(ClpThis->ClpConfig))->ClpStage[nNumStage - 1];
  rA0 = ClpLastStage->rA[ClpLastStage->nMaxSize - 1];
  rV0 = ClpLastStage->rV[ClpLastStage->nMaxSize - 1];
  rAlt0 = ClpLastStage->rAlt[ClpLastStage->nMaxSize - 1];
}

/* Adept Outputs Altitudes, Cambridge Outputs Velocities */
/* Require Altitudes, Velocities, and Accelerations */
/* Overloaded to Fill In Blanks As Appropriate */
ClpThis->vFillinDataBlanks(ClpThis, rA0,  rV0, rAlt0);

ClpThis->vGetVMaxAMaxAMin(ClpThis);

/* Find Burnout Start */
if(!bRepeat) {
 (VOID) ClpThis->nFindBurnout(ClpThis);
}

/* For Lower Stages May Choose To Read In Thrust Curve */
/* Otherwise An Empty Function is Called */
ClpThis->vReadThrustCurve(ClpThis);


/* For Lower Stages May Choose To Read In Cd Curve */
/* Otherwise The DoNothing Function is Called */
ClpThis->vReadCdCurve(ClpThis);

/* Derive the Drag Parameters */
/* May Be Overloaded for Lower Stages */
ClpThis->vMakeShapeConstantsAndCds(ClpThis);

/* Back Out the Thrust Curve */
/* Iteratively Distributing Mass Over Thrust Time */
/* May Be Overloaded For Lower Stages */
rTotalImpulse = ClpThis->rMakeThrustCurve(ClpThis, rM, TRUE);
if (rTotalImpulse < 0.0) {
  bRepeat = TRUE;
  return (bRepeat);
}

rOldTotal = (REAL) 0;

i = 0;

while((i < 5)||(ABS(rTotalImpulse - rOldTotal)/rTotalImpulse > EPSILON)) {

   rOldTotal = rTotalImpulse;
   rTotalImpulse = ClpThis->rMakeThrustCurve(ClpThis, rM, FALSE);
   i++;

}

bRepeat = FALSE;
return(bRepeat);
}
/**************************************************************************/
VOID Stage_Cl__vDestructor(Stage_Clp ClpStage)
{
ClpStage->nMaxSize = (NINT) 0;
ClpStage->nStageNumber = (NINT) 0;
ClpStage->ClpConfig = NULL;
ClpStage->rRho = NULL;

/* Default Input is a Velocity Curve */
ClpStage->vFillinDataBlanks = NULL;

ClpStage->vMakeShapeConstantsAndCds = NULL;
ClpStage->rMakeSpaceVelocities = NULL;
ClpStage->rMakeThrustCurve =  NULL;
ClpStage->bAnalyzeStage =  NULL;
ClpStage->vReadCdCurve =  NULL;
ClpStage->rRho =  NULL;
ClpStage->vDestructor = NULL;
strcpy(ClpStage->caInputFileName, "");
strcpy(ClpStage->caInputCdCurveName, "");
strcpy(ClpStage->caInputFCurveName, "");
strcpy(ClpStage->caOutputFCurveName, "");
strcpy(ClpStage->caOutputCdCurveName, "");
return;
}
/**************************************************************************/
CBOOL Stage_Cl__bReadCambridgePrint(Stage_Clp ClpStage)
{
NINT i, nSignal;
FILE *ifp;
CBOOL MaxedOut;
REAL * rInputArray, rWork1, rWork2, rWork3, rWork4;
CHAR capInString[MAX_STRING];

   /*    ***** Read Velocity Data From Cambridge Hex Dump *****    */

rInputArray = ClpStage->rV;


ifp=fopen(ClpStage->caInputFileName,"r");
if (ifp == NULL)  return(FALSE);

MaxedOut = TRUE;
for(i=0; i< CAMBRIDGE_PRINT_PREAMBLE_SIZE; i++) {
    fgets(capInString, MAX_STRING, ifp);

}


/* Treatment of first reading: Assume an initial interval */
rInputArray[0] = 0.0;

   /*     ****** Read Actual Data ******     */
for(i=1; i< MAX_SIZE; i++) {

   if((nSignal = fscanf(ifp,"%f %f %f %f",
		 &rWork1, &rWork2, &rWork3, &rWork4)) == EOF) {
    MaxedOut = FALSE;
    break;
   }

   rInputArray[i] = (REAL) rWork3;

   /* Cambridge Accelerometers Multiply Velocity by 50 */
   rInputArray[i] *= ((REAL) (5280.0/3600.0) / MTOF);

}

fclose(ifp);

ClpStage->nMaxSize = i;

if(MaxedOut) {
  DISPLAY("Cambridge Data Set Size Maxed Out\n");
}

return(!MaxedOut);
}
/**************************************************************************/
CBOOL Stage_Cl__bReadCambridgeDump(Stage_Clp ClpStage)
{
NINT i, nWork, nSignal;
FILE *ifp;
CBOOL MaxedOut;
REAL * rInputArray;

   /*    ***** Read Velocity Data From Cambridge Hex Dump *****    */

rInputArray = ClpStage->rV;


ifp=fopen(ClpStage->caInputFileName,"r");
if (ifp == NULL)  return(FALSE);

MaxedOut = TRUE;


   /*     ****** Space Over Preamble Data ******     */
for(i=0; i< CAMBRIDGE_PREAMBLE_SIZE; i++) {
  if((nSignal = fscanf(ifp,"%x", &nWork)) == EOF)return(FALSE);
  if(i == CAMBRIDGE_INTERVAL_INDEX) {
    ClpStage->rDeltaT = ((REAL) (nWork + nWork)) * CAMBRIDGE_TICK_SIZE ;
  }
}

/* Treatment of first reading: Assume an initial interval */
rInputArray[0] = 0.0;

   /*     ****** Read Actual Data ******     */
for(i=1; i< MAX_SIZE; i++) {

   if(((nSignal = fscanf(ifp,"%x", &nWork)) == EOF)||
	 (nWork == (NINT) 0xffff)){
    MaxedOut = FALSE;
    break;
   }

   rInputArray[i] = (REAL) nWork;

   /* Cambridge Accelerometers Multiply Velocity by 50 */
   rInputArray[i] /= (REAL) 50.0;

}

fclose(ifp);

ClpStage->nMaxSize = i;

if(MaxedOut) {
  DISPLAY("Cambridge Data Set Size Maxed Out\n");
}

return(!MaxedOut);
}
/**************************************************************************/
CBOOL Stage_Cl__bReadAdeptData(Stage_Clp ClpStage){
NINT i, nSignal;
NINT nWork;
FILE *ifp;
CBOOL MaxedOut;
REAL * rInputArray;

rInputArray = ClpStage->rAlt;

ifp=fopen(ClpStage->caInputFileName,"r");
if (ifp == NULL)  return(FALSE);


/*   ***** Read Altitude Data From Adept Recording Altimeter *****   */

MaxedOut = TRUE;
for(i=0; i< MAX_SIZE; i++) {

   if((nSignal = fscanf(ifp,"%i", &nWork)) == EOF){
     MaxedOut = FALSE;
     break;
   }

   rInputArray[i] = (REAL) nWork;
   rInputArray[i] /= MTOF;

}

fclose(ifp);
ClpStage->nMaxSize = i;

if(MaxedOut) {
  DISPLAY("Adept Data Set Size Maxed Out\n");
}

return(!MaxedOut);
}
/**************************************************************************/
CBOOL Stage_Cl__bReadAccelerationData(Stage_Clp ClpStage){
NINT i, nWork, nSignal;
FILE *ifp;
CBOOL MaxedOut;
REAL * rInputArray;

rInputArray = ClpStage->rA;

ifp=fopen(ClpStage->caInputFileName,"r");
if (ifp == NULL)  return(FALSE);


/*   ***** Read Generic Acceleration Data - Placeholder Routine *****   */

MaxedOut = TRUE;
for(i=0; i< MAX_SIZE; i++) {

   if((nSignal = fscanf(ifp,"%f", &nWork)) == EOF){
     MaxedOut = FALSE;
     break;
   }

   rInputArray[i] = (REAL) nWork;
   rInputArray[i] /= MTOF;

}

fclose(ifp);
ClpStage->nMaxSize = i;

if(MaxedOut) {
  DISPLAY("Adept Data Set Size Maxed Out\n");
}

return(!MaxedOut);
}
/**************************************************************************/
VOID Stage_Cl__vSaveFlightFile(Stage_Clp ClpStage, FILE * ofp)
{
NINT i, nWork;
STATIC REAL rTime;
REAL rFactor;

rFactor = (REAL) 1.0 / ((Config_Clp)ClpStage->ClpConfig)->rAlt0Factor;


if(ClpStage->nStageNumber == 0) {
   rTime = (REAL) 0.0;
}

for(i = 0; i < ClpStage->nMaxSize; i++) {
   DISPLAYF(ofp, "%f    %f         %f       %f     %f\n", rTime,
      ClpStage->rAlt[i] * rFactor,
      ClpStage->rV[i] * rFactor,
      ClpStage->rA[i] * rFactor,
      ClpStage->rCdTable[(NINT)(ClpStage->rV[i] + (REAL) 0.5)]);

   rTime += ClpStage->rDeltaT;

}

}
/**************************************************************************/
VOID Config_Cl__vSaveFlightFile(Config_Clp ClpConfig)
{
NINT i, nWork;
FILE *ofp;
REAL rWork;
CHAR * capHeaders[3] =
{"# Time     Altitude(m)     Velocity(m/s)     Accel(m/s^2)     Cd",
 "# Time     Altitude(ft)    Velocity(ft/s)    Accel(ft/s^2)    Cd",
 "# Time     Altitude(ft)    Velocity(ft/s)    Accel(ft/s^2)    Cd"};

ClpConfig->caOutputFlightFileName[0] = '\0';

DISPLAY("Enter Output Flight File Name ");

GetLine( FALSE, ClpConfig->caOutputFlightFileName);

if(!strcmp(ClpConfig->caOutputFlightFileName,"")) {

  DISPLAY("No Flight File Saved \n");
  return;

}

ofp=fopen(ClpConfig->caOutputFlightFileName,"w");
if (ofp == NULL)  return;

ClpConfig->vPrintoutConfiguration(ClpConfig, ofp);

DISPLAYF(ofp, "%s\n", capHeaders[ClpConfig->nUnits - (NINT) 1]);

for(i=0; i < ClpConfig->nNumStages; i++) {

   ClpConfig->ClpStage[i]->vSaveFlightFile(ClpConfig->ClpStage[i], ofp);
}

fclose(ofp);
return;
}
/**************************************************************************/
CBOOL Stage_Cl__bReadInputKeyboard(Stage_Clp ClpStage)
{
NINT i, nWork;
FILE *ifp;
Config_Clp ClpConfig;

ClpConfig = (Config_Clp) ClpStage->ClpConfig;

 while(((ClpStage->nMachCorrectSwitch =
	(NINT) Prompt("Add Simulated Mach to Cd Table ? (1=Yes, 2=No) ",
	 (REAL) ClpStage->nMachCorrectSwitch)) != (NINT) 1) &&
	 (ClpStage->nMachCorrectSwitch != 2));
/*
 ClpStage->rMaxCd = Prompt("Maximum Low-Velocity Cd ", ClpStage->rMaxCd);
*/

ClpStage->rMaxCd = 999.0;

ClpStage->rLaunchMassInput =
	Prompt(ClpConfig->caLaunchMassLabel, ClpStage->rLaunchMassInput);

ClpStage->rLaunchMass = ClpStage->rLaunchMassInput
			      * ClpConfig->rLaunchMassFactor;

ClpStage->rMassPropInput =
	Prompt(ClpConfig->caMassPropLabel, ClpStage->rMassPropInput);

ClpStage->rMassProp = ClpStage->rMassPropInput
			      * ClpConfig->rMassPropFactor;

ClpStage->rMaxDiamInput =
	Prompt(ClpConfig->caMaxDiameterLabel, ClpStage->rMaxDiamInput);

ClpStage->rMaxDiam = ClpStage->rMaxDiamInput
			      * ClpConfig->rMaxDiameterFactor;

ClpStage->rAx = PI * ClpStage->rMaxDiam * ClpStage->rMaxDiam / (REAL) 4.0;

if(ClpStage->nDataKind != CAMBRIDGE) {
  ClpStage->rDeltaT =
	Prompt("Enter Time Interval in Seconds ", ClpStage->rDeltaT);
}

DISPLAY("Enter Input File Name ");
GetLine( FALSE, ClpStage->caInputFileName);


return(ClpStage->bReadInputData(ClpStage));
}
/**************************************************************************/
VOID Stage_Cl__vSaveCdCurve(Stage_Clp ClpStage)
{
NINT i, nWork;
FILE *ofp;
REAL rWork;

ClpStage->caOutputCdCurveName[0] = '\0';

DISPLAY("Enter Output Cd File Name for Stage %i ", ClpStage->nStageNumber +1);

GetLine( FALSE, ClpStage->caOutputCdCurveName);

if(!strcmp(ClpStage->caOutputCdCurveName,"")) {

  DISPLAY("No Cd File Saved \n");
  return;

}

ofp=fopen(ClpStage->caOutputCdCurveName,"w");
if (ofp == NULL)  return;

   rWork = ClpStage->rMaxDiam * ((REAL) 100.0);

   fprintf(ofp,"#%s\n", ClpStage->caOutputCdCurveName);
   fprintf(ofp,"%f %f %f %f\n", rWork, rWork, rWork, rWork);
   fprintf(ofp,"#%f\n", rWork);
   fprintf(ofp,"#%f\n", rWork);
   fprintf(ofp,"#%f\n", rWork);

for(i=0; i< ClpStage->nCdTableSize; i++) {
  /* Extend Cd Curve - dumbly for the moment */
  if(ClpStage->rCdTable[i] > (REAL) 0.0) {
    rWork = ClpStage->rCdTable[i];
  }

   fprintf(ofp,"%f\n", rWork);

}

fclose(ofp);
return;
}
/**************************************************************************/
VOID Stage_Cl__vSaveFCurve(Stage_Clp ClpStage)
{
NINT i, nMaxSize;
REAL rTime = (REAL) 0.0;
FILE *ofp;

ClpStage->caOutputFCurveName[0] = '\0';

DISPLAY("Enter Output Thrust File Name for Stage %i ",
		  ClpStage->nStageNumber + (NINT) 1);

GetLine( FALSE, ClpStage->caOutputFCurveName);

if(!strcmp(ClpStage->caOutputFCurveName,"")) {

  DISPLAY("No Thrust File Saved \n");
  return;

}

ofp=fopen(ClpStage->caOutputFCurveName,"w");
if (ofp == NULL)  return;

fprintf(ofp,"%s\n", ClpStage->caOutputFCurveName);

nMaxSize = ClpStage->nStartCoastPhase;
fprintf(ofp,"%f  %f  %f  %i\n", ClpStage->rTotalImpulse,
(REAL) 1000.0 * ClpStage->rMassProp,
(REAL) (ClpStage->nStartCoastPhase - (NINT) 1)*
			   ClpStage->rDeltaT,
nMaxSize - (NINT) 1);


nMaxSize = ClpStage->nStartCoastPhase;

for(i=0; i< nMaxSize; i++) {
   fprintf(ofp,"%f    %f\n", rTime, ClpStage->rThrust[i]);
   rTime += ClpStage->rDeltaT;
}

fclose(ofp);
return;
}
/**************************************************************************/
VOID Stage_Cl__vPrintoutStage(Stage_Clp ClpThis, NINT nStageNumber,
				  FILE * ofp)
{
Config_Clp ClpConfig;
CHAR *capYesNo[2] = {"Yes", "No"};
CHAR *capBrandLabel[4] = { " Cambridge Dump", "Cambridge Print",
			      "Generic", "Adept"};

ClpConfig = (Config_Clp) ClpThis->ClpConfig;

DISPLAYF(ofp, "#Stage Number = %i\n", nStageNumber + (NINT) 1);

DISPLAYF(ofp,"#Kind of Input for stage: %s \n",
	capBrandLabel[ClpThis->nDataKind - (NINT) 1]);

DISPLAYF(ofp, "#Add Simulated Mach to Cd Table ? : %s \n#",
	 capYesNo[ClpThis->nMachCorrectSwitch- (NINT) 1]);

DISPLAYF(ofp, "Maximum Low-Velocity Cd: %f \n#", ClpThis->rMaxCd);

DISPLAYF(ofp, ClpConfig->caLaunchMassLabel);
DISPLAYF(ofp, ":  %f\n#", ClpThis->rLaunchMassInput);

DISPLAYF(ofp, ClpConfig->caMassPropLabel);
DISPLAYF(ofp, ":  %f\n#",ClpThis->rMassPropInput);

DISPLAYF(ofp, ClpConfig->caMaxDiameterLabel);
DISPLAYF(ofp, ":  %f\n" ,ClpThis->rMaxDiamInput);

DISPLAYF(ofp, "#Time Interval in Seconds %f\n", ClpThis->rDeltaT);

DISPLAYF(ofp, "#Input File Name  %s\n", ClpThis->caInputFileName);

DISPLAYF(ofp, "#\n");

/* ** End Of Outputted Input Data ** */


DISPLAYF(ofp,"#Total Impulse = %f ns", ClpThis->rTotalImpulse);
DISPLAYF(ofp,"           %f lb Sec\n", ClpThis->rTotalImpulse/((REAL) 4.45));



DISPLAYF(ofp, "#Max Altitude = %f Meters",
			     ClpThis->rAlt[ClpThis->nMaxSize - (NINT)1]);
DISPLAYF(ofp, "        %f Feet\n",  MTOF *
			     ClpThis->rAlt[ClpThis->nMaxSize - (NINT)1]);

#ifdef NEVER_DEFINED
DISPLAYF(ofp, "#Cutoff Velocity = %f Meters/Sec ",
			   ClpThis->rV[ClpThis->nStartCoastPhase - 1]);

DISPLAYF(ofp, "   %f Feet/Sec \n", MTOF
			   ClpThis->rV[ClpThis->nStartCoastPhase - 1]);
#endif

DISPLAYF(ofp, "#Max Velocity = %f Meters/Sec ", ClpThis->rVMax);
DISPLAYF(ofp, "    %f Feet/Sec \n", MTOF * ClpThis->rVMax);

DISPLAYF(ofp, "#Max Acceleration = %f gs", (ClpThis->rAMax / G));
DISPLAYF(ofp, "       Min = %f gs\n", (ClpThis->rAMin / G));

DISPLAYF(ofp, "#\n#Isp = %f Sec ", ClpThis->rIsp);
DISPLAYF(ofp, "          Max Mach Number = %f\n",
    ClpThis->rVMax / GetMach1(((Config_Clp) ClpThis->ClpConfig)->rTempK));

DISPLAYF(ofp, "#Flight Time = %f Sec", (REAL) ClpThis->nMaxSize *
				      ClpThis->rDeltaT);
DISPLAYF(ofp, "     Burning Time = %f Sec\n#\n",
			   (REAL) (ClpThis->nStartCoastPhase - (NINT) 1)*
			   ClpThis->rDeltaT);

DISPLAYF(ofp, "#Last Velocity Registered in Coast Phase = %f m/sec\n",
	     ClpThis->rLastV);

return;
}
/**************************************************************************/
VOID Config_Cl__vPrintoutConfiguration(Config_Clp ClpConfig, FILE * ofp)
{
NINT i;

CHAR *capSharpRounded[2] = {"SHARP", "Rounded"};
CHAR * capUnitsLabel[3] = {"Metric", "English", "Combination"};


if(ofp == STDOUT) {
  CLEAR_SCREEN;
}

DISPLAYF(ofp,"#Units are: %s\n",
      capUnitsLabel[ClpConfig->nUnits - (NINT) 1]);

DISPLAYF(ofp,"#");

DISPLAYF(ofp, ClpConfig->caAlt0Label);
DISPLAYF(ofp,": %f\n#", ClpConfig->rAlt0Input);

DISPLAYF(ofp, ClpConfig->caTempLabel);
DISPLAYF(ofp,": %f\n", ClpConfig->rTempInput);
DISPLAYF(ofp,"#Nose Type:  %s\n",
    capSharpRounded[ClpConfig->nNoseSwitch - (NINT) 1]);

DISPLAYF(ofp,"#\n");

for(i=0; i < ClpConfig->nNumStages; i++) {

   ClpConfig->ClpStage[i]->vPrintoutStage(ClpConfig->ClpStage[i], i, ofp);

}
}
/**************************************************************************/
VOID Stage_Cl__vConstructor(Stage_Clp ClpStage, NINT nStageNumber,
			    Config_Clp ClpConfig)
{

ClpStage->nMaxSize = MAX_SIZE;
ClpStage->nCdTableSize = CD_TABLE_SIZE;

ClpStage->nStageNumber = nStageNumber;
ClpStage->ClpConfig = ClpConfig;
ClpStage->rRho = Stage_Cl__rRho;
ClpStage->bReadInputKeyboard = Stage_Cl__bReadInputKeyboard;
ClpStage->vPrintoutStage = Stage_Cl__vPrintoutStage;
ClpStage->vSaveFlightFile = Stage_Cl__vSaveFlightFile;

/* Default Input is a Velocity Curve */
ClpStage->nFindBurnout = Stage_Cl__nFindBurnout;
ClpStage->vFillinDataBlanks = Stage_Cl__vMakeAccsAndAlts;
ClpStage->rRho = Stage_Cl__rRho;

ClpStage->vConvolve = Stage_Cl__vConvolve;
ClpStage->vSmooth = Stage_Cl__vExponentialSmoothing;

ClpStage->vAssertMaximumLowVCd =
		   Stage_Cl__vAssertMaximumLowVCd;

ClpStage->rMaxCd = (REAL) 3.45;

ClpStage->vGetVMaxAMaxAMin = Stage_Cl__vGetVMaxAMaxAMin;


ClpStage->vMakeShapeConstantsAndCds = Stage_Cl__vMakeShapeConstantsAndCds;
ClpStage->rMakeSpaceVelocities = Stage_Cl__rMakeSpaceVelocities;
ClpStage->rMakeThrustCurve = Stage_Cl__rMakeThrustCurve;
ClpStage->bAnalyzeStage = Stage_Cl__bAnalyzeStage;

/* Reserved For MultiStage Version */
ClpStage->vReadThrustCurve = Stage_Cl__vDoNothing;
ClpStage->vReadCdCurve = Stage_Cl__vDoNothing;
ClpStage->vSaveFCurve = Stage_Cl__vSaveFCurve;
ClpStage->vSaveCdCurve = Stage_Cl__vSaveCdCurve;


ClpStage->vDestructor = Stage_Cl__vDestructor;
   strcpy(ClpStage->caInputFileName, "");
   strcpy(ClpStage->caInputCdCurveName, "");
   strcpy(ClpStage->caInputFCurveName, "");
   strcpy(ClpStage->caOutputFCurveName, "");
   strcpy(ClpStage->caOutputCdCurveName, "");




return;
}
/**************************************************************************/
VOID Config_Cl__vAnalyzeConfiguration(Config_Clp ClpThis)
{
NINT i;
for (i = 0; i < ClpThis->nNumStages; i++) {

   while (ClpThis->ClpStage[i]->bAnalyzeStage(ClpThis->ClpStage[i], i));
}
return;
}
/**************************************************************************/
VOID Config_Cl__vSaveConfigurationData(Config_Clp ClpConfig)
{
NINT i;

ClpConfig->vSaveFlightFile(ClpConfig);

for(i = 0; i < ClpConfig->nNumStages; i++) {
   ClpConfig->ClStage[i].vSaveCdCurve(ClpConfig->ClpStage[i]);
   ClpConfig->ClStage[i].vSaveFCurve(ClpConfig->ClpStage[i]);
}
return;
}
/**************************************************************************/
CBOOL Config_Cl__bInputConfiguration(Config_Clp ClpThis)
{
STATIC NINT nTimesThrough = 0;
NINT i;
Stage_Clp ClpStage;
CBOOL bRetCode = TRUE;
CHAR cReply;


if(nTimesThrough > 0) {
   while(((cReply = GetReply("Another Run? (Y/N)")) != 'N') &&
	 (cReply != 'Y'));

   if(cReply != 'Y')return(FALSE);
}

nTimesThrough++;


CLEAR_SCREEN;

while (((ClpThis->nUnits =
	(NINT) Prompt("Units (1=Metric, 2=English, 3=Combination ",
		  ClpThis->nUnits)) != 1) &&
		  (ClpThis->nUnits != 2) &&
		  (ClpThis->nUnits != 3));

/* Multiple Stages Not Implemented In This Version */
#ifdef MULTIPLE_STAGES
while((ClpThis->nNumStages = (NINT) Prompt("How Many Stages? ",
			      (REAL) ClpThis->nNumStages) )> MAX_STAGE) {
CLEAR_SCREEN;
}
#else

ClpThis->nNumStages = 1;

#endif

   switch(ClpThis->nUnits) {

   case METRIC :
      ClpThis->rMaxDiameterFactor = (REAL) 0.01;
      ClpThis->rMassPropFactor = (REAL) 0.001;
      ClpThis->rLaunchMassFactor = (REAL) 0.001;
      ClpThis->rAlt0Factor = (REAL) 1.0;

      strcpy(ClpThis->caTempLabel, "Enter Launch Temperature (C) ");
      strcpy(ClpThis->caMaxDiameterLabel, "Enter Maximum Diameter (cm) ");
      strcpy(ClpThis->caMassPropLabel, "Enter Propellant Mass (g) ");
      strcpy(ClpThis->caLaunchMassLabel, "Enter Launch Mass (g) ");
      strcpy(ClpThis->caAlt0Label, "Enter Launch Site Altitude (m) ");
      break;

   case ENGLISH :
      ClpThis->rMaxDiameterFactor = (REAL) 1.0 /(((REAL) 12.0)* (MTOF));
      ClpThis->rMassPropFactor = OZTOKG;
      ClpThis->rLaunchMassFactor = OZTOKG;
      ClpThis->rAlt0Factor = (REAL) 1.0 / (MTOF);

      strcpy(ClpThis->caTempLabel, "Enter Launch Temperature (F) ");
      strcpy(ClpThis->caMaxDiameterLabel, "Enter Maximum Diameter (in) ");
      strcpy(ClpThis->caMassPropLabel, "Enter Propellant Mass (oz) ");
      strcpy(ClpThis->caLaunchMassLabel, "Enter Launch Mass (oz) ");
      strcpy(ClpThis->caAlt0Label, "Enter Launch Site Altitude (ft) ");
      break;

   case COMBO :
      ClpThis->rMaxDiameterFactor = (REAL) 0.01;
      ClpThis->rMassPropFactor = (REAL) 0.001;
      ClpThis->rLaunchMassFactor = (REAL) 0.001;
      ClpThis->rAlt0Factor = (REAL) 1.0 / (MTOF);

      strcpy(ClpThis->caTempLabel, "Enter Launch Temperature (F) ");
      strcpy(ClpThis->caMaxDiameterLabel, "Enter Maximum Diameter (cm) ");
      strcpy(ClpThis->caMassPropLabel, "Enter Propellant Mass (g) ");
      strcpy(ClpThis->caLaunchMassLabel, "Enter Launch Mass (g) ");
      strcpy(ClpThis->caAlt0Label, "Enter Launch Site Altitude (ft) ");
      break;
   }


ClpThis->rAlt0Input =
	Prompt(ClpThis->caAlt0Label, ClpThis->rAlt0Input);

ClpThis->rAlt0 = ClpThis->rAlt0Input * ClpThis->rAlt0Factor;

ClpThis->rTempInput =
	Prompt(ClpThis->caTempLabel, ClpThis->rTempInput);

if(ClpThis->nUnits != METRIC){

  ClpThis->rTempF = ClpThis->rTempInput;

  ClpThis->rTempC =
  ((REAL) 5.0) * (ClpThis->rTempF - (REAL) 32.0)/((REAL) 9.0);
}
else {

  ClpThis->rTempC = ClpThis->rTempInput;

  ClpThis->rTempF =
   ((REAL) 9.0) * (ClpThis->rTempF)/((REAL) 5.0) + (REAL) 32.0;

}

ClpThis->rTempK = ClpThis->rTempC + KELVIN0;


while (((ClpThis->nNoseSwitch =
	(NINT) Prompt("Nose Type (1=Sharp, 2=Rounded) ",
		  (REAL) ClpThis->nNoseSwitch)) != 1) &&
		  (ClpThis->nNoseSwitch != 2));

DISPLAY("\n");

for (i = 0; i < ClpThis->nNumStages; i++) {
   ClpStage = &ClpThis->ClStage[i];
   DISPLAY("Kind of Input for stage %i \n", i + (NINT) 1);

   while(((ClpStage->nDataKind =(NINT)
    Prompt("1=Cambridge Dump, 2=Cambridge Print, 3= Acceleration ",
	       ClpStage->nDataKind))!= CAMBRIDGE) &&
	       (ClpStage->nDataKind != CAMBRIDGEP) &&
	       (ClpStage->nDataKind != ACCELERATION));

     switch(ClpStage->nDataKind) {
     case ADEPT:
     ClpStage->bReadInputData = Stage_Cl__bReadAdeptData;
     bRetCode = ClpStage->bReadInputKeyboard(ClpStage);
     ClpStage->vFillinDataBlanks = Stage_Cl__vMakeVelsAndAccs;

			     break;

     case CAMBRIDGE:

       ClpStage->bReadInputData = Stage_Cl__bReadCambridgeDump;
       bRetCode = ClpStage->bReadInputKeyboard (ClpStage);
       ClpStage->vFillinDataBlanks = Stage_Cl__vMakeAccsAndAlts;
			     break;

     case CAMBRIDGEP:

       ClpStage->bReadInputData = Stage_Cl__bReadCambridgePrint;
       bRetCode = ClpStage->bReadInputKeyboard (ClpStage);
       ClpStage->vFillinDataBlanks = Stage_Cl__vMakeAccsAndAlts;
			     break;

     case ACCELERATION:
       ClpStage->bReadInputData = Stage_Cl__bReadAccelerationData;
       bRetCode = ClpStage->bReadInputKeyboard (ClpStage);
       ClpStage->vFillinDataBlanks = Stage_Cl__vMakeVelsAndAlts;
		       break;

     default:          break;
     }/*Switch */
   }

return(bRetCode);
}
/**************************************************************************/

VOID Config_Cl__vDestructor(Config_Clp ClpConfig)
{
NINT i;

   for(i=0; i < MAX_STAGE; i++) {
     Stage_Cl__vDestructor(&ClpConfig->ClStage[i]);
   }

   ClpConfig->bInputConfiguration = NULL;

   ClpConfig->vAnalyzeConfiguration = NULL;

   ClpConfig->vPrintoutConfiguration = NULL;

   ClpConfig->vSaveConfigurationData = NULL;

   ClpConfig->vDestructor = NULL;




return;
}
/**************************************************************************/
CBOOL Config_Cl__bConstructor(Config_Clp ClpConfig)
{
CBOOL bRetCode = TRUE;
NINT i;

   memset(ClpConfig, 0, sizeof(Config_Cl));

   ClpConfig->vDestructor = Config_Cl__vDestructor;

   ClpConfig->bInputConfiguration = Config_Cl__bInputConfiguration;

   ClpConfig->vAnalyzeConfiguration = Config_Cl__vAnalyzeConfiguration;

   ClpConfig->vSaveFlightFile = Config_Cl__vSaveFlightFile;

   ClpConfig->vPrintoutConfiguration = Config_Cl__vPrintoutConfiguration;

   ClpConfig->vSaveConfigurationData = Config_Cl__vSaveConfigurationData;

   if(bRetCode) {
      for(i=0; i < MAX_STAGE; i++) {
	 ClpConfig->ClpStage[i] = &ClpConfig->ClStage[i];
	 Stage_Cl__vConstructor(ClpConfig->ClpStage[i], i, ClpConfig);
      }
   }

return(bRetCode);
}
/**************************************************************************/
VOID main()
{
Config_Cl ClConfig;
Config_Clp ClpConfig = &ClConfig;


if(!Config_Cl__bConstructor(ClpConfig)) {
  DISPLAY("Unsuccessful Construction of Configuration Class");
  exit(99);
}

vBanner();

CLEAR_SCREEN;

while (ClpConfig->bInputConfiguration(ClpConfig)) {

   ClpConfig->vAnalyzeConfiguration(ClpConfig);

   ClpConfig->vPrintoutConfiguration(ClpConfig, STDOUT);

   ClpConfig->vSaveConfigurationData(ClpConfig);
}

/* Unnecessary but clean */
ClpConfig->vDestructor(ClpConfig);

return;
}
